"""
Module handles making requests to NewsAPI and then processing
the data recieved from NewsAPI
"""
import json
import requests

def news_API_request(covid_terms="Covid COVID-19 coronavirus")->list:
    """
    Function makes a request to NewsAPI.
    It then processes this data to find articles containing the terms
    given.
    It returns a list of dictionaries. Each dictionary is an article
    """
    with open('config.json',encoding="utf-8") as file:
        config_data=json.load(file)

        api_key=config_data['APIKey']

    lowered_covid_terms=covid_terms.lower()
    term_list=lowered_covid_terms.split()
    #Requests top headlines from the UK from NewsAPI
    url =('https://newsapi.org/v2/top-headlines?'
        'country=gb&'
        'apiKey='+api_key)

    response = requests.get(url)

    news=response.json()['articles']

    covid_articles=[]

    #Iterates through the articles received and finds
    #those which contain the key terms
    for article in news:
        if any(x in article['title'].lower() for x in term_list):
            covid_articles.append(article)

        elif article['description'] is not None:
            if any(x in article['description'].lower() for x in term_list):
                covid_articles.append(article)

    return covid_articles

def update_news(news_to_add:list,removed_news=[])->list:
    """
    Function takes a list of dictionaries as is returned by
    news_API_request. These articles are then checked against
    a list of previously removed articles and any that match
    are removed. Also if there are too many articles some are
    removed.
    A list of dictionaries is returned.
    """
    if isinstance(news_to_add,list):
        #Compares new API response against previously deleted news
        removed_news_title=[]
        for news in removed_news:
            removed_news_title.append(news['title'])

        for article in news_to_add:
            if article['content'] is None or article['title'] in removed_news_title:
                news_to_add.remove(article)

        #Removes article in length is 4 or less
        while len(news_to_add)>4:
            del news_to_add[-1]

        return news_to_add

    return None



"""This module handles all functions and procedures
related to dealing with fetching and porcessing covid
data from both csv files and the covid API
"""
import csv
import sched
import time
from datetime import datetime
from uk_covid19 import Cov19API

s=sched.scheduler(time.time, time.sleep)

def parse_csv_data(csv_filename:str)->list:
    """
    Function takes a csv filename and proccesses the data.
    The function returns a list of strings with each string
    being a seperate row from the csv file.
    """

    with open(csv_filename,'r',encoding="utf-8") as file:
        data=csv.reader(file)
        rows = []
        for row in data:
            #Creates a string in which values are seperated by commas
            day=row[0]+","+row[1]+","+row[2]+","+row[3]+","+row[4]+","+row[5]+","+row[6]
            rows.append(day)

    return rows

def process_covid_csv_data(covid_csv_data):
    """
    Function takes csv data in either a list or dictionary form.
    If the data comes in list form then non useful columns are
    removed. If the data is a dictionary then a list is created
    from the dictionary values
    Then the data is extracted and returned as integers
    Note: when calculating last 7 days cases a day is ignored
    if it has less than half the previous day's cases as this
    would be a significant drop in cases suggesting a more likely
    reason is simply incomplete data.
    """
    useful_data=[]
    if isinstance(covid_csv_data,list):
        #Eliminating column headers and columns that don't contain useful data
        for i in range(0,len(covid_csv_data)):
            row=covid_csv_data[i].split(',')
            del row[:4]
            useful_data.append(row)

        del useful_data[:1]

    elif isinstance(covid_csv_data,dict):
        useful_data=list(covid_csv_data.values())

    #Finding Cumulative Deaths
    found=False
    cum_deaths=0
    row=0
    while found is False and row<len(useful_data):
        if useful_data[row][0]=='' or useful_data[row][0] is None:
            row+=1

        else:
            cum_deaths=int(useful_data[row][0])
            found=True

    #Finding Hospital Cases
    found=False
    hospital_cases=0
    row=0
    while found is False and row<len(useful_data):
        if useful_data[row][1]=='' or useful_data[row][1] is None:
            row+=1

        else:
            hospital_cases=int(useful_data[row][1])
            found=True

    #Finding Last 7 Day Cases
    found=False
    last_7_days=0
    row=0
    #While loop will find the first value of valid case data
    while found is False and row<len(useful_data):
        #Will ignore data if reported cases are less than half of previous day
        if useful_data[row][2]=='' or \
           useful_data[row][2] is None or \
           int(useful_data[row][2])<int(useful_data[row+1][2])*0.5:
            row+=1

        else:
            found=True

    #Accumulates 7 days of data starting at the day with first valid data
    for i in range(row,row+7):
        last_7_days=last_7_days+int(useful_data[i][2])

    return last_7_days,hospital_cases,cum_deaths

def covid_API_request(location="Exeter",location_type="ltla")->dict:
    """
    Function sends a request to the UK covid API
    This data recieved is then processed into a dictionary
    with dates as keys and a list of data as values
    """
    #Use UK covid interactive map to easily find location names
    location=['areaType='+location_type,'areaName='+location]

    cases_and_deaths = {
        "Date": "date",
        "CumDeaths": "cumDailyNsoDeathsByDeathDate",
        "HospitalCases": "hospitalCases",
        "NewCases": "newCasesBySpecimenDate",
    }

    api=Cov19API(filters=location, structure=cases_and_deaths)

    data=api.get_json()

    unparsed_dict=data["data"]
    parsed_dict={}

    #Formats data as dictionary with dates as keys.
    #Values are strings formatted the same as the extracted csv data
    for i in range(0,len(unparsed_dict)):
        values=[unparsed_dict[i]["CumDeaths"],\
                unparsed_dict[i]["HospitalCases"],\
                unparsed_dict[i]["NewCases"]]
        parsed_dict[unparsed_dict[i]["Date"]]=values

    return parsed_dict

def schedule_time_calculator(update_interval:str)->int:
    """
    Function recieves a time string in HH:MM form and then returns
    how many seconds until the time is next reached
    """
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    interval=0
    #current_time_sec and update_sec are how many seconds from 00:00 the time is.
    current_time_sec=int(current_time[0:2])*60*60+int(current_time[3:5])*60+int(current_time[6:8])
    update_sec=int(update_interval[0:2])*60*60+int(update_interval[3:5])*60
    #If the delta is positive it means that the scheduled time is for the current day
    #If it's negative it needs to be scheduled tomorrow
    delta=update_sec-current_time_sec

    if delta>0:
        interval=delta

    elif delta<0:
        interval=(24*60*60-current_time_sec)+update_sec

    return interval

def schedule_covid_updates(update_interval,update_name):
    """
    Function recieves a time in seconds and then schedules
    a covid update in n seconds.
    """
    s.enter(update_interval,1,process_covid_csv_data,argument=(covid_API_request("Exeter","ltla"),))
    s.run()



"""Imports modules from covid_data_handler and covid_news_handling to
to create the user interface using flask to run a development server
which runs the index.html script
"""
import logging
import time
import sched
import json
from flask import Flask
from flask import render_template
from flask import request
from covid_data_handler import covid_API_request
from covid_data_handler import schedule_time_calculator
from covid_data_handler import process_covid_csv_data
from covid_news_handling import news_API_request
from covid_news_handling import update_news

#Flask has inbuilt logger called werkzeug
logging.getLogger('werkzeug').setLevel(logging.INFO)
logging.basicConfig(filename='logging.log', encoding='utf-8',
level=logging.WARNING,format=\
'%(asctime)s - %(module)s - %(funcName)s - %(lineno)s - %(name)s - %(levelname)s - %(message)s')

with open('config.json',encoding="utf-8") as file:
    config_data=json.load(file)

app=Flask(__name__)
s=sched.scheduler(time.time, time.sleep)
location=config_data['location']
nation=config_data['nation']
area_type=config_data['areaType']
national_7_day_infections,national_hospital,national_cum_deaths,local_7_day_infections=0,0,0,0
updates=[]
removed_news=[]

try:
    local_7_day_infections=process_covid_csv_data(covid_API_request(location,area_type))[0]
    national_7_day_infections,national_hospital,national_cum_deaths=\
    process_covid_csv_data(covid_API_request(nation,"nation"))
except Exception as error:
    logging.error(error)

try:
    news_articles=update_news(news_API_request(),removed_news)
except Exception as error:
    logging.error(error)

def update_covid_data(update: dict):
    """Procedure takes in an update as dictionary
    If the update is in the global list updates it calls the
    covid_API_request and process_covid_csv_data functions
    """
    try:
        #Checks update is in the update array meaning if update is deleted it is cancelled
        if update in updates:
            local_7_day_infections=process_covid_csv_data(covid_API_request(location,area_type))[0]
            national_7_day_infections,national_hospital,national_cum_deaths=\
            process_covid_csv_data(covid_API_request(nation,"nation"))
    except Exception as error:
        logging.error(error)

def news_updater(update: dict):
    """Procedure takes in an update as dictionary.
    If the update is in the global list updates it calls the
    news_API_request and update_news functions
    """
    try:
        #Checks update is in the update array meaning if update is deleted it is cancelled
        if update in updates:
            news_articles=update_news(news_API_request(),removed_news)
    except Exception as error:
        logging.error(error)

def update_creator(name: str)->dict:
    """Takes the name of an update to be created as string.
    The function then fetches all the other nessacary update
    information using flask request module.
    Finally creates a dictionary with this information which
    is appended to global list updates. The dictionary update
    is returned
    """
    repeat=request.args.get('repeat')
    update_time=request.args.get('update')
    covid_data=request.args.get('covid-data')
    news=request.args.get('news')

    content=update_time+" - "
    if repeat=='repeat':
        content="Repeats at "+content

    if covid_data=='covid-data' and news=='news':
        content=content+"Update covid data and news articles"

    elif covid_data=='covid-data':
        content=content+"Update covid data "

    elif news=='news':
        content=content+"Update news"

    update={'title':name,
            'content':content,
            'time':update_time,
            'repeat':repeat,
            'covidData':covid_data,
            'news':news
            }

    updates.append(update)

    return update

def update_remover(update: dict):
    """Takes an update to delete as a dictionary.
    Then removes the update from global list updates
    """
    if update in updates:
        updates.remove(update)

def news_remover(title_to_delete: str):
    """Takes the title of an article to delete as a string.
    Then removes the article from global list news_articles"
    """
    for article in news_articles:
        if article['title']==title_to_delete:
            removed_news.append(article)
            news_articles.remove(article)

def update_scheduler(update: dict):
    """Procedure takes the update to be scheduled as a dictionary.
    The values of this dictionary are then checked to see what
    updates need to be scheduled depending on what options have
    been selected. If the update repeats the procedure will call
    itself
    """
    if update in updates:
        interval=schedule_time_calculator(update['time'])

        if update['covidData']=='covid-data':
            s.enter(interval,1,update_covid_data,(update,))

        if update['news']=='news':
            s.enter(interval,1,news_updater,(update,))

        if update['repeat']=='repeat':
#If the update repeats, on completion of one update the update is passed into the scheduler again
            s.enter(interval,1,update_scheduler,(update,))

        else:
            #If the update doesn't repeat it's removed when completed
            s.enter(interval,1,update_remover,(update,))

@app.route('/index')
def hello():
    """Function which acts on the /index path.
    If an input is submitted by the user from the html
    script different functions and procedures are triggered
    to handle this input.
    Function returns a list of information to be rendered
    by html script.
    """
    s.run(blocking=False)
    update_name=request.args.get('two')
    if update_name:
        update_scheduler(update_creator(update_name))
    update_to_delete=request.args.get('update_item')
    if update_to_delete:
        for update in updates:
            if update['title']==update_to_delete:
                update_remover(update)

    news_to_delete=request.args.get('notif')
    if news_to_delete:
        news_remover(news_to_delete)

    return render_template('index.html',
                           location=location,
                           local_7day_infections=local_7_day_infections,
                           nation_location=nation,
                           national_7day_infections=national_7_day_infections,
                           hospital_cases="Hospital cases: "+str(national_hospital),
                           deaths_total="Total deaths: "+str(national_cum_deaths),
                           news_articles=news_articles,
                           updates=updates,
                           image="image.jpg",
                           title='Covid Updates'
                           )


if __name__ == '__main__':
    app.run()



import setuptools

with open("README.md", "r") as fh:
    long_description=fh.read()

setuptools.setup(
    name="dashboard-pkg-mtrenchard",
    version="1.0",
    author="Matt Trenchard",
    author_email="mjt235@exeter.ac.uk",
    description="A covid dashboard providing the latest Covid information. To be used for demonstration.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/m4ttT10/2021-Covid-Dashboard-CA",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)
